#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
//#include <cuda_runtime.h>
#include "cublas_v2.h"
#include "cuda-diagnose.h"
#include "timing418.h"

//--------------------------------------------------------------------------
// When comparing two single precision floating point numbers, what is
// close enough to be considered equality?
#define CLOSE_ENOUGH 1e-3

// Convenient allocation macro.
#define NewArray(Type, N) ((Type *)(malloc(N*sizeof(Type))))

// Fortran column major indexing.  We don't actually need the number
// of columns (n), but we'll include it as an argument anyway for when
// we do C-style (row major) indexing.
#define IDX2F(i,j,m,n) (((j)*(m))+(i))

// C-style row major indexing.  We don't actually need the number
// of rows (m), but we'll include it as an argument anyway for when
// we do Fortran-style (column major) indexing.
//#define IDX2C(i,j,m,n) (((i)*(n))+(j))

//--------------------------------------------------------------------------
// Generates a matrix with random entries between 0 and 1 (inclusive).
void rand_matrix(float *A, uint m, uint n) {
  // We use rows on the inner loop to be efficient in a column-major
  // indexing scheme.  Column major is the Fortran convention used by
  // BLAS (and is opposite the usual C row-major convention).
  for(uint j = 0; j < n; j++)
    for(uint i = 0; i < m; i++)
      A[IDX2F(i,j,m,n)] = (double)rand() / (double)RAND_MAX;
}

//----------------------------------------------------------------------
void compare_matrices(const float *C1, const float *C2, 
		      const uint m, const uint p, 
		      const char *s1, const char *s2) {

  for(uint i = 0; i < m; i++)
    for(uint k = 0; k < p; k++)
      if(fabs(C1[IDX2F(i,k,m,p)] - C2[IDX2F(i,k,m,p)]) > CLOSE_ENOUGH) {
	fprintf(stderr, "\nERROR: i=%d, k=%d, %s=%f, %s=%f\n",
		i, k, s1, C1[IDX2F(i,k,m,p)], s2, C2[IDX2F(i,k,m,p)]);
	exit(-1);
      }
}

//----------------------------------------------------------------------
void s418mm_cpu(float *A, float *B, float *C, 
		uint m, uint n, uint p, uint ntrials) {

  for(uint trial = 0; trial < ntrials; trial++)
    // Basic matrix multiplication code.  It will not be particularly
    // efficient since it was not coded in a cache aware fashion.
    for(uint i = 0; i < m; i++)
      for(uint k = 0; k < p; k++) {
	float sum = 0.0;
	for(uint j = 0; j < n; j++)
	  sum += A[IDX2F(i,j,m,n)] * B[IDX2F(j,k,n,p)];
	C[IDX2F(i,k,m,p)] = sum;
      }
}


//----------------------------------------------------------------------
__global__ void s418mm_kernel(float *A, float *B, float *C,
			      uint m, uint n, uint p) {

  // Determine our location in the matrix using nvcc built-ins.
  uint i = blockIdx.x*blockDim.x + threadIdx.x;
  uint k = blockIdx.y*blockDim.y + threadIdx.y;

  // If we are not an extra thread.
  if((i < m) && (k < p)) {
    // Calculate a single element of the output matrix.
    float sum = 0.0;
    for(uint j = 0; j < n; j++)
      sum += A[IDX2F(i,j,m,n)] * B[IDX2F(j,k,n,p)];
    C[IDX2F(i,k,m,p)] = sum;
  }
}

//----------------------------------------------------------------------
void s418mm_gpu(float *A, float *B, float *C, 
		uint m, uint n, uint p, uint ntrials) {
  float *dev_A, *dev_B, *dev_C;
  int size_A = m * n * sizeof(float);
  int size_B = n * p * sizeof(float);
  int size_C = m * p * sizeof(float);

  // Allocate memory on the GPU for all of the arrays.
  cudaTry(cudaMalloc((void**)(&dev_A), size_A));
  cudaTry(cudaMalloc((void**)(&dev_B), size_B));
  cudaTry(cudaMalloc((void**)(&dev_C), size_C));
  
  // Copy over the input arrays.
  cudaTry(cudaMemcpy(dev_A, A, size_A, cudaMemcpyHostToDevice));
  cudaTry(cudaMemcpy(dev_B, B, size_B, cudaMemcpyHostToDevice));

  // Set up the computational grid.  We will use maximum size square
  // blocks, but your choice may have been different.
  dim3 dimGrid(ceil(m/32.0), ceil(p/32.0), 1);
  dim3 dimBlock(32, 32, 1);

  // Calculate the matrix multiplication.
  for(uint trial = 0; trial < ntrials; trial++)
    s418mm_kernel<<<dimGrid, dimBlock>>>(dev_A, dev_B, dev_C, m, n, p);

  // Copy back the result.
  cudaTry(cudaMemcpy(C, dev_C, size_C, cudaMemcpyDeviceToHost));

  // Free up the arrays on the GPU.
  cudaTry(cudaFree(dev_A));
  cudaTry(cudaFree(dev_B));
  cudaTry(cudaFree(dev_C));
}

//----------------------------------------------------------------------
// Note that the cuBLAS handle must be created in advance (we don't
// want to include the time to create the handle since it apparently
// takes a long time at least on the first call).
void s418mm_cublas(float *A, float *B, float *C, 
		   uint m, uint n, uint p, uint ntrials, 
		   cublasHandle_t handle) {
  float *dev_A, *dev_B, *dev_C;
  int size_A = m * n * sizeof(float);
  int size_B = n * p * sizeof(float);
  int size_C = m * p * sizeof(float);

  // Allocate memory on the GPU for all of the arrays.
  cudaTry(cudaMalloc((void**)(&dev_A), size_A));
  cudaTry(cudaMalloc((void**)(&dev_B), size_B));
  cudaTry(cudaMalloc((void**)(&dev_C), size_C));

  // Put the arrays onto the GPU using cuBLAS.
  cublasTry(cublasSetMatrix(m, n, sizeof(*A), A, m, dev_A, m));
  cublasTry(cublasSetMatrix(n, p, sizeof(*B), B, n, dev_B, n));

  // Perform the matrix multiplication.
  float alpha = 1.0f;
  float beta = 0.0f;
  for(uint trial = 0; trial < ntrials; trial++)
    // Note strange ordering of matrix size parameters: <rows of A &
    // C>, <columns of B & C>, <columns of A and rows of B>.
    cublasTry(cublasSgemm(handle, CUBLAS_OP_N, CUBLAS_OP_N, m, p, n, 
			  &alpha, dev_A, m, dev_B, n, &beta, dev_C, m));

  // Copy the final result back (also ensures that the GPU has
  // finished its computation).
  cublasTry(cublasGetMatrix(m, p, sizeof(*C), dev_C, m, C, m));

  // Free up the arrays on the GPU.
  cudaTry(cudaFree(dev_A));
  cudaTry(cudaFree(dev_B));
  cudaTry(cudaFree(dev_C));
}

//----------------------------------------------------------------------
void s418mm_test(uint m, uint n, uint p, uint ntrials) {
  float *A, *B, *C_cpu, *C_gpu, *C_cublas;
  timer_418 t0, t1;
  cublasHandle_t handle;

  // Set up cuBLAS.
  cublasTry(cublasCreate(&handle));

  // Allocate and fill the matrices.
  A = NewArray(float, m*n);
  rand_matrix(A, m, n);
  B = NewArray(float, n*p);
  rand_matrix(B, n, p);
  C_cpu = NewArray(float, m*p);
  C_gpu = NewArray(float, m*p);
  C_cublas = NewArray(float, m*p);
  
  // Time the calculations.
  timestamp(&t0);
  s418mm_cpu(A, B, C_cpu, m, n, p, ntrials);
  timestamp(&t1);
  double t_cpu = elapsed_seconds(t1, t0);
  printf("CPU: %10.3e\n", t_cpu / ntrials);

  timestamp(&t0);
  s418mm_gpu(A, B, C_gpu, m, n, p, ntrials);
  timestamp(&t1);
  double t_gpu = elapsed_seconds(t1, t0);
  printf("GPU: %10.3e\n", t_gpu / ntrials);

  timestamp(&t0);
  s418mm_cublas(A, B, C_cublas, m, n, p, ntrials, handle);
  timestamp(&t1);
  double t_cublas = elapsed_seconds(t1, t0);
  printf("cuBLAS: %10.3e\n", t_cublas / ntrials);

  // Check the calculations.
  compare_matrices(C_cpu, C_gpu, m, p, "cpu", "gpu");
  compare_matrices(C_cpu, C_cublas, m, p, "cpu", "cuBLAS");  
  
  // Free up all the resources.
  free(A);
  free(B);
  free(C_cpu);
  free(C_gpu);
  free(C_cublas);
  cublasDestroy(handle);
}


//----------------------------------------------------------------------
int main(int argc, char **argv) {
  if(argc < 2) {
    fprintf(stderr, "usage: mmult-compare m [n] [p] [ntrials]\n");
    exit(1);
  }

  uint m, n, p, ntrials;
  char *conversion_garbage;

  // Parse the input arguments.
  m = strtoul(argv[1], &conversion_garbage, 10);
  if((m < 1) || (strlen(conversion_garbage) > 0)) {
    fprintf(stderr, "mmult-compare, unable to convert %s to positive integer m\n", argv[1]);
    exit(1);
  }

  if(argc > 2) {
    n = strtoul(argv[2], &conversion_garbage, 10);
    if((n < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "mmult-compare, unable to convert %s to positive integer n\n", argv[2]);
      exit(1);
    }
  }
  else
    n = m;

  if(argc > 3) {
    p = strtoul(argv[3], &conversion_garbage, 10);
    if((p < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "mmult-compare, unable to convert %s to positive integer p\n", argv[3]);
      exit(1);
    }
  }
  else
    p = n;

  if(argc > 4) {
    ntrials = strtoul(argv[4], &conversion_garbage, 10);
    if((ntrials < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "mmult-compare, unable to convert %s to positive integer ntrials\n", argv[4]);
      exit(1);
    }
  }
  else
    ntrials = 1;

  if(argc > 5) {
    fprintf(stderr, "usage: mmult-compare m [n] [p] [ntrials]\n");
    exit(1);
  }

  // Uncomment the call to srand() if you want different random number
  // sequences on each run.
  // srand((uint)time(NULL));

  // Run the tests.
  printf("Running with m=%d, n=%d, p=%d and %d trials.\n", m, n, p, ntrials);
  s418mm_test(m, n, p, ntrials);
  printf("\n");
}
